module Main where
 
import Regexp
import ConstructorMonad(LblM,getNewLbl,initLbl,foldlM,returnLbl,thenLbl,thenLbl_)
import System

{-
  retester takes a file containing a number of regexp test examples.
  They should be of the form:
  pattern subject result repl expected
  where
   pattern - a regexp pattern
   subject - what to match it against
   result - whether it matched "y", failed to match "n, or was illegal "c"
   repl - what to take from the match result eg ${m} for the whole match
   expected - what was the expected actual match.
   
  The above are all separated by tabs (\t).

  The results will be :
    how many tests there were : eg "1 to 263 tests" 
    Test num illegal - if the regexp was found to be illegal but should
                       not have been, for the regexp on line num
    Test num for "pat" against "subject" failed gave #mres# should be #expect#
    yesresult actresult -
        if test failed, where:
         num - line number
         pat - regexp pattern
         subject - what matched against
         mres - is actual result of match
         expect - is what was expected
         yesresult - is whether should have matched (Bool)
         actresult - is whether did match (Bool) 
    
    Invalid num data length -
        if test data had invalid number of arguments
          num - line number
          data - is all the test data on that line
          length - the number of arguments in that line of test data
-}
       


main =  getArgs >>= \argv -> 
       case argv of
         (inp:out:[]) -> 
                         readFile inp >>= \contents -> 
                         writeFile out (acc contents) 

         _ -> putStr usage_msg 

usage_msg :: String
usage_msg = unlines 
            ["Usage: retester infile outfile",
             "where",
             "  infile - is file with regexp test examples",
             "  outfile - is file produced with test results"] 
acc = runtest.lines

split :: (a -> Bool)	-- what to split list at 
      -> [a]		-- the list to split
      -> [[a]]		-- the list once split
split p [] = []
split p (x:xs) | p x = w1:split p s1''
               | otherwise = w:split p s''
          where
                   (w1,s1'') = break p xs
                   (w,s'') = break p (x:xs)

runtest :: [String]	-- the list of test regexps 
        -> String	-- the result 
runtest file = let (res,length) = initLbl (foldlM testone "" file) 1
               in
               "1 to "++(show length)++" tests\n"
               ++ res

testone :: String -> String -> LblM String
testone one out = test (split (== '\t') one) `thenLbl` \res ->
                  returnLbl (out++res)

test :: [String] -> LblM String

test (pat1:subject:result:repl:expect:[]) 
  = let (pat,flags) = case pat1 of
                       ('\'':xs) -> ((init.init) xs,[Case_Insensitive])
                       xs -> (xs,[])
        matchresult =  searchExtS pat flags [] [af1,af2] subject
        mres = replaceS repl matchresult
        newresult = result == "y"
        illegal = result == "c"
    in
       getNewLbl `thenLbl` \name ->
       if (not (legalRegexp matchresult)) && illegal then
         returnLbl("")
       else if (not (legalRegexp matchresult)) then
         returnLbl("Test " ++ show name ++ " illegal\n")
       else if not (expect == mres && newresult == matchedAny matchresult) then
         returnLbl(
         "Test" ++ show name ++ " for \"" ++ pat ++ "\" against \"" ++ subject 
         ++ "\" failed\n"++
         "gave #" ++ mres ++ "#should be #" ++ expect ++ "#" ++ show newresult 
         ++ show (matchedAny matchresult) ++ "\n\n")
       else
         returnLbl("")--"Test " ++ show name ++ "\n")
 
test xs =
     getNewLbl `thenLbl` \name ->
     returnLbl("Invalid "++show name++" "++show xs++" "++show (length xs)++"\n")
 
                 

af1 _ ('c':xs) = True
af1 _ _ = False

af2 _ (x:xs) = x == 'c' || x == 'd'
af2 _ _ = False

mf1 ('b':'o':'b':xs) = Just ("bob",xs)
mf1 _ = Nothing





handleReference :: (Show a,Read a) => 
                   REmatch a -> String -> Maybe ([a],[Char]) 
handleReference rem repl =
       let matchres = ref repl
       in 
         if not (matchedAny matchres) then
            Nothing
         else if (matchedSubexp matchres "whole") then -- matched ${m}
            Just (wholeMatch rem, afterMatch matchres)
         else if matchedSubexp matchres "bef" then -- matched ${m_before_}
            Just (beforeMatch rem, afterMatch matchres)
         else if matchedSubexp matchres "aft" then -- matched ${m_after_}
            Just (afterMatch rem,afterMatch matchres)
         else --if matchedSubexp matchres "sub" then -- matched ${..}...
            Just (subexpMatch rem (subexpMatch matchres "sub"),
                  afterMatch matchres)
 
{-
replaceS - uses the REmatch given, and a string of replacement info to
           build the replacement list.
           eg if ${m} = "ab"
              replace "${m}d" rem = "abd"
-}
replaceS replacement rem = replace' replacement rem
  where
     replacer = searchS "\\\\ el@(.)|\\$" [] 
     replace' repl rem =
        let matchres = replacer repl
        in
          if not (matchedAny matchres) then
             repl
          else if matchedSubexp matchres "el" then
             beforeMatch matchres ++ subexpMatch matchres "el" 
             ++ replace' (afterMatch matchres) rem
          else
             beforeMatch matchres ++ 
             (case handleReference rem (afterMatch matchres) of
               Nothing -> replace' (afterMatch matchres) rem
               (Just (ms,as)) -> ms ++ replace' as rem) 
     
ref = searchS "^\\{( bef@(m_before_)| aft@(m_after_)| whole@(m)| sub@(\\w+))\\}"
 []

